/*
* V:1.4
* Anbindung des Raspberry PI ber Modbus mittels modbus.h und wiringPI.h
* Auslesen 1-Wire DS18B20 ber Module "w1-gpio" "w1-therm" und deren Verzeichnis
*/
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <modbus/modbus.h>
#include <wiringPi.h>
#include <string.h>
#include <time.h>

modbus_t *ctx;
modbus_mapping_t *mb_mapping;
int socket,temp;
char tempsensor[20];

int gpio[] = {7,0,1,2,3,4,5,6,8,9,10,11,12,13,14,15};	//16-IO Adressen
//Umrechnung Bit in GPIO-Port unter https://projects.drogon.net/raspberry-pi/wiringpi/pins/

FILE *ptr_file;
char buf[100];
char *p;
double tempx;
	
/*
typedef struct {
  int nb_bits;
  int nb_input_bits;
  int nb_input_registers;
  int nb_registers;
  uint8_t *tab_bits;
  uint8_t *tab_input_bits;
  uint16_t *tab_input_registers;
  uint16_t *tab_registers;
} modbus_mapping_t;
*/


int mb_init(void)
{
  printf ("modbus init\n") ;
  ctx = modbus_new_tcp("0.0.0.0", 1502);
  /* modbus_set_debug(ctx, TRUE); */
  mb_mapping = modbus_mapping_new(8, 16, 0, 32);
  /*  Aufbau:
      mb_mapping = modbus_mapping_new(
        UT_BITS_ADDRESS + UT_BITS_NB,
        UT_INPUT_BITS_ADDRESS + UT_INPUT_BITS_NB,
        UT_REGISTERS_ADDRESS + UT_REGISTERS_NB,
        UT_INPUT_REGISTERS_ADDRESS + UT_INPUT_REGISTERS_NB);
  */
  if (mb_mapping == NULL) 
  {
    fprintf(stderr, "Failed to allocate the mapping: %i %s\n", errno, modbus_strerror(errno));
    modbus_free(ctx);
    return -1;
  }
  
  printf ("waiting for connection!\n") ;
  socket = modbus_tcp_listen(ctx, 1);
  modbus_tcp_accept(ctx, &socket);
  printf ("connected!\n") ;
    return 1;
}

int mb_close(void)
{
  close(socket);
  modbus_mapping_free(mb_mapping);
  modbus_close(ctx);
  modbus_free(ctx);
    return 1;
}

int temp_onewire (char tempsensor[])	//Liest abgespeicherte 1-Wire Werte aus 
{										//hierfr mssen Module "w1-gpio" "w1-therm" 
	char dateiname[60];					//in etc/modules gestartet werden 
	strcpy(dateiname,"/sys/bus/w1/devices/");
	strcat(dateiname,tempsensor);
	strcat(dateiname,"/w1_slave");
	ptr_file = fopen(dateiname, "r");
	if (!ptr_file)
        return 3333;						//bei Fehler unsinnigen Tempwert zurckgeben
	while (fgets(buf,100, ptr_file)!=NULL)
	{};//printf("%s",buf);
	p = strstr(buf, "t=") + 2;
	p[strlen(p)-1] = 0;			//remove new line char
	//tempx = atof(p)/1000;		//cast to float	
	tempx = atof(p)/100;		//cast to float
	tempx+= 0.5;				//Ziel auf eine Nachkommastelle zu runden 
	temp = (int)tempx; 			//Division durch 10 erfolgt erst in IPS - bertragung als int
	fclose(ptr_file);
	return temp;
}
	
int gpio_init(void)
{
  int n;
  printf ("GPIO Init\n") ;
  if (wiringPiSetup () == -1)
    return 1 ;
  /*set GPIOs to OUTPUT*/
  for(n=0;n<8;n++) pinMode(gpio[n], OUTPUT);	//GPIO4=1-Wire in IPS Adressen 0-7 1bit 
  /*set GPIOs to INTPUT*/
  for(n=8;n<16;n++) pinMode(gpio[n], INPUT);	//GPIO4=1-Wire in IPS Adressen 8-15 1bit nur lesen
  printf ("Init Done!\n") ;
  return 1;
}


int mb_poll(void)
{
  uint8_t query[MODBUS_TCP_MAX_ADU_LENGTH];
  int rc;
  int n;
  rc = modbus_receive(ctx, query); 
  if (rc != -1) 
  {
	for(n=8;n<=16;n++)								//Input-Ports abfragen
	{
		mb_mapping->tab_input_bits[n]=digitalRead(n);
	} 
	time_t t;
	t = time(NULL); 
	int sek=t%10;				//durch 10=alle10Sek
	if (sek == 0)   {      		//durch 10=alle10Sek
		for(n=16;n<18;n++)		//Anzahl Sensoren-jedes Verzeichnis einzeln auslesen	16-da ModbusAdr=16 fr ersten Sensor
		{																				// jeder Sensor 16bit Word in IPS
			switch (n)	{
				case 16:	strcpy(tempsensor,"28-000004ab2720");	//jeweiliger DS18B20 Sensor
						break;
				case 17:	strcpy(tempsensor,"0815");				//jeweiliger DS18B20 Sensor
						break;
				default: break;
			}	
			temp=temp_onewire(tempsensor);
			if(temp!=3333) mb_mapping->tab_input_registers[n]=temp;			
		}		
	} 	
	modbus_reply(ctx, query, rc, mb_mapping);
	for(n=0;n<8;n++) digitalWrite(gpio[n], mb_mapping->tab_bits[n]);	
    return 1;
  } 
  else 
  {
    /* Connection closed by the client or error */
    return -1;
  }
}


int poll_loop(void)
{
  for(;;) 
  {
    if(mb_poll()==-1)
    {
      printf("mb error \n");
      mb_close();
      break;
    } 
  }
  printf("quitting loop: %i %s\n", errno, modbus_strerror(errno));
  if(errno == 104) //connection closed by peer
  {
    if(mb_init()) poll_loop();
  }
 return 1;
}

int main(void)
{
  gpio_init();
  if(mb_init()) poll_loop();
  return 0;
}

